from django import template
from django.conf import settings
from django.utils.translation import get_language
from django.utils.safestring import mark_safe

from ..blocks import BlockProcessor

register = template.Library()

@register.simple_tag
def multilang(val):
    return val.get(get_language(), val.get(settings.LANGUAGE_CODE))


@register.simple_tag
def blocks(page_blocks):
    return mark_safe(BlockProcessor().render(page_blocks))

@register.simple_tag
def block_scripts(page_blocks):
    return mark_safe(BlockProcessor().render_script_tags(page_blocks))

@register.simple_tag
def block_stylesheets(page_blocks):
    return mark_safe(BlockProcessor().render_stylesheet_tags(page_blocks))


def get_blocks_for_page(page):
    qs = page.get_blocks_for_language(get_language())
    if qs.count() == 0:
        qs = page.get_blocks_for_language(settings.LANGUAGE_CODE)
    return qs

@register.simple_tag
def pageblocks(page):
    return blocks(get_blocks_for_page(page))

@register.simple_tag
def pageblocks_scripts(page):
    return block_scripts(get_blocks_for_page(page))

@register.simple_tag
def pageblocks_stylesheets(page):
    return block_stylesheets(get_blocks_for_page(page))



# Generated by Django 3.2.5 on 2022-03-16 11:50

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('pageblocks', '0003_auto_20211107_1804'),
    ]

    operations = [
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('image', models.ImageField(upload_to='pageblocks/%Y/%m/%d/')),
            ],
        ),
    ]



# Generated by Django 3.2.5 on 2021-10-12 16:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pageblocks', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='pageblock',
            name='content',
        ),
        migrations.AddField(
            model_name='pageblock',
            name='data',
            field=models.JSONField(default=dict),
        ),
    ]



import base64
import copy
import imghdr
import re
import uuid

from django.conf import settings
from django.core.exceptions import ValidationError
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy, gettext
from django.core.files.base import ContentFile

from .utils import class_from_name
from .models import Image


class BaseField(object):
    input_type = 'text'
    input_classes = []

    def __init__(self, label=None, required=False, additional_classes=None, *args, **kwargs):
        self.label = label
        self.required = required

        if additional_classes:
            self.input_classes.update(additional_classes)

    def serialize_field_definition(self, id):
        return {
            'required': self.required,
            'input_type': self.input_type,
            'label': str(self.label) if self.label else self.id,
            'class': ' '.join(self.input_classes)
        }
        
class CharField(BaseField):
    input_type = 'text'

class TextField(BaseField):
    input_type = 'textarea'

class HTMLField(TextField):
    input_classes = ['html']

class ImageField(BaseField):
    input_type = 'image'

class BlockStreamField(BaseField):
    input_type = 'blockstream'

    def serialize_field_definition(self, id):
        fd = super().serialize_field_definition(id)
        fd['initial'] = []
        return fd


class BlockValidationError(Exception):
    def __init__(self, msg, field_id):
        super().__init__(msg)
        self.field_id = field_id


class BlockRenderingError(Exception):
    pass


class BlockProcessor(object):
    def blocks_to_representation(self, blocks, data=None):
        if not data:
            data = []

        for page_block in blocks.order_by('index'):
            block_class = class_from_name(page_block.type)
            block = block_class(data=page_block.data, instance=page_block)

            data.append({
                'id': str(page_block.id),
                'type': page_block.type,
                'data': block.data_to_representation()
            })
        return data

    def clean(self, data, parent_indexes=[], language=''):
        if not data:
            return data

        if not language:
            raise Exception('No language')

        for index, block_data in enumerate(data):
            block_class = class_from_name(block_data['type'])
            block = block_class(block_data['data'])
            try:
                block_data['data'] = block.clean(parent_indexes=parent_indexes + [index], language=language)
            except BlockValidationError as bve:
                raise ValidationError(f'B:{language}:{self.format_index(parent_indexes, index)}:{bve.field_id}:' + str(bve))

        return data

    def format_index(self, parent_indexes, index):
        indexes = parent_indexes + [index]
        return ','.join([str(i) for i in indexes])

    def save(self, page, data, parent=None, language=None):
        processed_blocks = []
        
        for block_index, block_data in enumerate(data):
            block_class = class_from_name(block_data['type'])
            block = block_class(block_data,
                                instance=page.blocks.model.objects.get(page=page, id=block_data['id']) if block_data.get('id', None) else None)

            processed_blocks += block.save(page=page,
                                           language=language if not parent else parent.language,
                                           block_index=block_index, parent=parent)

        if not parent:
            cleanup_qs = page.blocks.model.objects.filter(page=page, language=language).exclude(id__in=[str(rec.id) for rec in processed_blocks])
            cleanup_qs.delete()

        return processed_blocks

    def render(self, blocks):
        return ''.join([block.get_block().render() for block in blocks])

    def render_script_tags(self, blocks):
        scripts = []
        for block in blocks:
            for script in block.get_block().get_scripts():
                if not script:
                    continue

                if script[0] != '<':
                    script = '<script type="text/javascript" src="%s"></script>' % script
                scripts.append(script)
        return '\n'.join([script for script in set(scripts)])

    def render_stylesheet_tags(self, blocks):
        stylesheets = []
        for block in blocks:
            for ss in block.get_block().get_stylesheets():
                if not ss:
                    continue

                if ss[0] != '<':
                    ss = '<link href="%s" rel="stylesheet" />' % ss
                stylesheets.append(ss)

        return '\n'.join([ss for ss in set(stylesheets)])


class BaseBlock(object):
    template_name = None
    name = None
    description = None
    fields = ()

    def __init__(self, data=None, instance=None, *args, **kwargs):
        self.data = data
        self.instance = instance

    @classmethod
    def serialize_field_definitions(cls):
        return {
            id: field.serialize_field_definition(id) for id, field in cls.fields
        }

    def data_to_representation(self):
        return self.data

    def data_to_internal_value(self, data):
        return data

    def clean(self, *args, **kwargs):
        """ Validate the block data """
        for field_id, field in self.fields:
            if field.required and not self.data.get(field_id, None):
                raise BlockValidationError(gettext('This field is required'), field_id=field_id)

        return self.data

    def get_instance_for_saving(self, page, language, block_index, parent, *args, **kwargs):
        instance = self.instance if self.instance else page.blocks.model(page=page, language=language)
        instance.language = language
        instance.type = self.data['type']
        instance.data = self.data_to_internal_value(self.data['data'])
        instance.index = block_index
        instance.parent = parent
        return instance

    def save(self, page, language, block_index, parent, *args, **kwargs):
        self.instance = self.get_instance_for_saving(page, language, block_index, parent, *args, **kwargs)
        self.instance.save()
        return [self.instance]

    def render(self):
        if not self.template_name:
            raise BlockRenderingError(gettext('No template_name defined for') + '.'.join([self.__class__.__module__, self.__class__.__name__]))
        
        return render_to_string(self.template_name, self.get_render_context_data())

    def get_render_context_data(self, *args, **kwargs):
        return {
            'instance': self.instance,
            'block': self.data_to_representation()
        }

    def get_scripts(self, *args, **kwargs):
        """
        Script dependencies to include for this block
        """
        return []

    def get_stylesheets(self, *args, **kwargs):
        """
        Stylesheet dependencies to include for this block
        """
        return []


class HTMLBlock(BaseBlock):
    """
    Renders the content as raw HTML
    """
    template_name = 'pageblocks/blocks/html.html'
    name = gettext_lazy('HTML')
    description = gettext_lazy('Raw HTML content')
    fields = (
        ('html', HTMLField(label=gettext_lazy('Content'), required=True)),
    )

class ImageBlock(BaseBlock):
    template_name = 'pageblocks/blocks/image.html'
    name = gettext_lazy('Image')
    description = gettext_lazy('A simple image')
    fields = (
        ('image', ImageField(label=gettext_lazy('Image'), required=True)),
        ('alt', CharField(label=gettext_lazy('Alt text'), required=False)),
        ('class', CharField(label=gettext_lazy('Class'), required=False)),
    )

    def data_to_representation(self):
        data = self.data
        if data.get('image_id', None):
            try:
                data['image'] = Image.objects.get(id=data['image_id']).image.url
            except Image.DoesNotExist:
                pass
        return data

    def data_to_internal_value(self, data):
        image = None
        if self.instance and self.instance.data.get('image_id', None):
            image = Image.objects.filter(id=self.instance.data['image_id']).first()

        if not re.search('^\/|^(?i)http', data['image']):
            if not image:
                image = Image()

            if image.image:
                image.image.delete()

            image.image = self.image_to_content_file(data['image'])
            image.save()

        data['image_id'] = str(image.id) if image else None

        del data['image']
        return data

    def image_to_content_file(self, data):
        if 'data:' in data and ';base64,' in data:
            header, data = data.split(';base64,')

        # Try to decode the file. Return validation error if it fails.
        try:
            decoded_file = base64.b64decode(data)
        except Exception:
            raise TypeError('Invalid image data')

        file_name = str(uuid.uuid4())[:12]
        file_extension = self.get_file_extension(file_name, decoded_file)

        return ContentFile(decoded_file, name="%s.%s" % (file_name, file_extension))

    def get_file_extension(self, file_name, decoded_file):
        extension = imghdr.what(file_name, decoded_file)
        extension = "jpg" if extension == "jpeg" else extension
        return extension


class ContainerBlock(BaseBlock):
    """
    A special block type that contains sub-blocks (useful for defining layouts such as columns)
    """
    template_name = 'pageblocks/blocks/container.html'
    name = gettext_lazy('Container')
    description = gettext_lazy('A container that contains other blocks')
    fields = (
        ('class', CharField(label=gettext_lazy('Class'), required=False)),
        ('blocks', BlockStreamField(label=gettext_lazy('Blocks'), required=True))
    )

    def clean(self, parent_indexes=[], language='', *args, **kwargs):
        data = super().clean(*args, **kwargs)
        BlockProcessor().clean(data['blocks'], parent_indexes=parent_indexes, language=language)
        return data

    def data_to_representation(self):
        data = super().data_to_representation()
        if self.instance:
            data['blocks'] = BlockProcessor().blocks_to_representation(self.instance.children.all())
        return data

    def save(self, page, language, block_index, parent, *args, **kwargs):
        if not self.instance:
            self.instance = page.blocks.model(page=page, language=language)

        block_data = copy.deepcopy(self.data['data'])
        sub_blocks = BlockProcessor().save(page, block_data.pop('blocks'), parent=self.instance)

        self.instance.language = language
        self.instance.type = self.data['type']
        self.instance.data = block_data
        self.instance.index = block_index
        self.instance.parent = parent
        self.instance.save()

        return [self.instance] + sub_blocks

    def get_render_context_data(self, *args, **kwargs):
        ctx = super().get_render_context_data(*args, **kwargs)
        ctx['blocks'] = self.instance.children.all().order_by('index')
        return ctx




import base64
import json

from django.conf import settings
from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import get_language, gettext
from django.utils.text import slugify

from .models import Page, PageBlock
from .utils import class_from_name
from .blocks import BlockProcessor


class LanguagesInputMixin(object):
    def get_context(self, *args, **kwargs):
        ctx = super().get_context(*args, **kwargs)
        ctx['languages'] = base64.b64encode(json.dumps({
            key: str(value) for key, value in settings.LANGUAGES
        }).encode()).decode()
        return ctx


class MultiLanguageInput(LanguagesInputMixin, forms.TextInput):
    template_name = 'admin/pageblocks/multi_language_input.html'


class PageBlockEditor(LanguagesInputMixin, forms.TextInput):
    template_name = 'admin/pageblocks/block_editor.html'

    def get_context(self, *args, **kwargs):
        ctx = super().get_context(*args, **kwargs)
        ctx['blocks'] = base64.b64encode(json.dumps({
            block_id: {
                'name': str(block_class.name),
                'description': str(block_class.description),
                'fields': block_class.serialize_field_definitions()
            } for block_id, block_class in Page.get_available_blocks()
        }).encode()).decode()
        ctx['translations'] = base64.b64encode(json.dumps({
            'labelBtnAdd': gettext('Add'),
            'labelUnknown': gettext('Unknown'),
            'labelConfirmRemoveBlock': gettext('Are you sure you want to remove this block?'),
            'labelConfirmCopy': gettext('Are you sure you want to do this?  This will overwrite all of the content in this page for this language.'),
            'labelCopyFrom': gettext('Copy from')
        }).encode()).decode()
        return ctx


class PageAdminForm(forms.ModelForm):
    blocks = forms.JSONField(widget=PageBlockEditor, initial=dict({
        key: [] for key, _ in settings.LANGUAGES
    }), required=False)
    slug = forms.SlugField(required=False)

    class Meta:
        model = Page
        fields = ('title', 'slug',)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.instance and self.instance.pk:
            self.init_blocks()

    def init_blocks(self):
        languages = [language for language, _ in settings.LANGUAGES]
        self.fields['blocks'].initial = {
            language: BlockProcessor().blocks_to_representation(self.instance.blocks.filter(parent=None, language=language)) for language in languages
        }

    def clean(self):
        data = self.cleaned_data
        if not data.get('slug', None) and data.get('title', {}):
            data['slug'] = self.generate_slug(data['title'].get(get_language(), gettext('Untitled')))
        return data

    def clean_blocks(self):
        """ Validate the data for each block (and recurse if it's a container block) """
        data = self.cleaned_data['blocks']
        if not data:
            return data

        languages = [language for language, _ in settings.LANGUAGES] + list(data.keys())

        return {
            language: BlockProcessor().clean(data[language], language=language) for language in set(languages)
        }

    def generate_slug(self, val):
        slug = slugify(val)
        offset = 0
        qs = Page.objects.all() if not self.instance and not self.instance.pk else Page.objects.exclude(pk=self.instance.pk)

        while qs.filter(slug=slug).count() > 0:
            offset += 1
            slug = '%s_%d' % (slugify(val), offset)

        return slug

    def save(self, commit=True):
        page = super().save(commit=commit)
        if commit:
            self.save_blocks(page)
        return page

    def save_blocks(self, page):
        block_data = self.cleaned_data.get('blocks', {})
        if not block_data:
            block_data = {}

        languages = [language for language, _ in settings.LANGUAGES] + list(block_data.keys())
        processed_blocks = []
        for language in set(languages):
            processed_blocks += BlockProcessor().save(page, block_data.get(language, []),
                                                      language=language)



# Generated by Django 3.2.5 on 2021-11-07 17:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('pageblocks', '0002_auto_20211012_1854'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='pageblock',
            options={'ordering': ['index']},
        ),
        migrations.AddField(
            model_name='pageblock',
            name='parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='pageblocks.pageblock'),
        ),
    ]



import uuid

from django.conf import settings
from django.db import models
from django.utils.translation import get_language, gettext

from .utils import class_from_name
from . import PAGEBLOCKS_DEFAULT_AVAILABLE


class MultiLanguageField(models.JSONField):
    def __init__(self, default=dict, *args, **kwargs):
        super().__init__(default=default, *args, **kwargs)

    def formfield(self, **kwargs):
        from .forms import MultiLanguageInput
        defaults = {'widget': MultiLanguageInput}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class AbstractPage(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = models.SlugField(unique=True, null=False, blank=False)
    title = MultiLanguageField()

    class Meta:
        abstract = True

    def __str__(self):
        return self.title.get(get_language(), gettext('Untitled'))

    @classmethod
    def get_available_block_type_classes(cls):
        try:
            return settings.PAGEBLOCKS_AVAILABLE_BLOCKS
        except AttributeError:
            return PAGEBLOCKS_DEFAULT_AVAILABLE

    @classmethod
    def get_available_blocks(cls):
        return [
            (c, class_from_name(c)) for c in cls.get_available_block_type_classes()
        ]

    def get_blocks_for_language(self, language):
        return self.blocks.filter(language=language, parent=None).order_by('index')

class Page(AbstractPage):
    pass

class AbstractPageBlock(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    language = models.CharField(max_length=2, db_index=True)
    index = models.IntegerField(default=-1)
    type = models.TextField()
    data = models.JSONField(default=dict)
    parent = models.ForeignKey('self', on_delete=models.CASCADE, blank=True, null=True,
                               related_name='children')

    class Meta:
        abstract = True
        ordering = ['index']

    def get_block(self):
        return class_from_name(self.type)(data=self.data, instance=self)


class PageBlock(AbstractPageBlock):
    page = models.ForeignKey(Page, on_delete=models.CASCADE, related_name='blocks', db_index=True)
    

class Image(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    image = models.ImageField(upload_to='pageblocks/%Y/%m/%d/')



import json

from django.test import TestCase, override_settings
from django.utils.translation import gettext_lazy, override as translation_override

from .models import Page, PageBlock
from .forms import PageAdminForm

from . import PAGEBLOCKS_DEFAULT_AVAILABLE

class AvailableBlockTestCase(TestCase):
    """
    Test logic to get available block types, with and without overridden settings
    """

    def test_without_settings_override(self):
        self.assertEqual(Page.get_available_block_type_classes(), PAGEBLOCKS_DEFAULT_AVAILABLE)

    @override_settings(PAGEBLOCKS_AVAILABLE_BLOCKS=[
        'pageblocks.blocks.RawHTMLBlock'
    ])
    def test_with_settings_override(self):
        self.assertEqual(Page.get_available_block_type_classes(), [
            'pageblocks.blocks.RawHTMLBlock'
        ])


@override_settings(LANGUAGES=[
    ('es', gettext_lazy('Spanish')),
    ('en', gettext_lazy('English')),
], LANGUAGE_CODE='en')
class PageEditingTestCase(TestCase):
    def test_create_empty_page(self):
        form = PageAdminForm(data={
            'title': {"es":"prueba", "en":"test"}
        })
        self.assertTrue(form.is_valid(), form.errors)
        with translation_override('en'):
            page = form.save()

        self.assertEqual(page.title['en'], 'test')
        self.assertEqual(page.title['es'], 'prueba')
        self.assertTrue(page.slug is not None)
        self.assertEqual(page.slug, 'test')

    def test_create_basic_page(self):
        form = PageAdminForm(data={
            'slug': 'basic_page',
            'title': {"es":"prueba", "en":"test"},
            'blocks': {"es":[{"type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>Es una prueba</b>"}},{"type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>Es una prueba otra vez</b>"}}],"en":[{"type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>This is a test</b>"}}]}
        })
        self.assertTrue(form.is_valid(), form.errors)
        with translation_override('en'):
            page = form.save()

        self.assertEqual(page.slug, 'basic_page')

        self.assertEqual(page.blocks.filter(language='en').count(), 1)
        self.assertEqual(page.blocks.filter(language='es').count(), 2)

        self.assertEqual(page.blocks.filter(language='en')[0].data['html'], '<b>This is a test</b>')
        self.assertEqual(page.blocks.filter(language='en')[0].index, 0)

        self.assertEqual(page.blocks.filter(language='es')[0].data['html'], '<b>Es una prueba</b>')
        self.assertEqual(page.blocks.filter(language='es')[0].index, 0)
        self.assertEqual(page.blocks.filter(language='es')[1].data['html'], '<b>Es una prueba otra vez</b>')
        self.assertEqual(page.blocks.filter(language='es')[1].index, 1)

    # TODO: Test editing a page .. does it load the blocks properly
    def test_editing_basic_page(self):
        PageBlock.objects.all().delete()

        form = PageAdminForm(data={
            'slug': 'basic_page',
            'title': {"es":"prueba", "en":"test"},
            'blocks': {"es":[{"type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>Es una prueba</b>"}},{"type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>Es una prueba otra vez</b>"}}],"en":[{"type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>This is a test</b>"}}]}
        })
        self.assertTrue(form.is_valid(), form.errors)
        with translation_override('en'):
            page = form.save()

        self.assertEqual(PageBlock.objects.all().count(), 3)

        form = PageAdminForm(instance=page)
        block_edit_data = {"es":[
            {"id": str(page.blocks.filter(language="es")[0].id), "type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>Es una prueba</b>"}},
            {"id": str(page.blocks.filter(language="es")[1].id), "type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>Es una prueba otra vez</b>"}}
        ],"en":
        [
            {"id": str(page.blocks.filter(language="en")[0].id), "type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>This is a test</b>"}}
        ]}
        self.assertDictEqual(form.fields['blocks'].initial, block_edit_data)

        block_edit_data['es'][0]['data']['html'] = '<b>He cambiado!</b>'
        del block_edit_data['es'][1]

        block_edit_data['en'].append(
            {"type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>This is a new block</b>"}}
        )

        form = PageAdminForm(instance=page, data={
            'slug': 'basic_page',
            'title': {"es":"prueba", "en":"test"},
            'blocks': block_edit_data
        })
        self.assertTrue(form.is_valid(), form.errors)

        page = form.save()

        self.assertEqual(page.blocks.filter(language='en').count(), 2)
        self.assertEqual(page.blocks.filter(language='es').count(), 1)

        self.assertEqual(page.blocks.filter(language='en')[0].data['html'], '<b>This is a test</b>')
        self.assertEqual(page.blocks.filter(language='en')[0].index, 0)
        self.assertEqual(page.blocks.filter(language='en')[1].data['html'], '<b>This is a new block</b>')
        self.assertEqual(page.blocks.filter(language='en')[1].index, 1)

        self.assertEqual(page.blocks.filter(language='es')[0].data['html'], '<b>He cambiado!</b>')
        self.assertEqual(page.blocks.filter(language='es')[0].index, 0)

    def test_create_page_missing_required_field(self):
        form = PageAdminForm(data={
            'slug': 'basic_page',
            'title': {"es":"prueba", "en":"test"},
            'blocks': {"es":[],"en":[
                {"type":"pageblocks.blocks.HTMLBlock","data":{}}
            ]}
        })
        with translation_override('en'):
            self.assertFalse(form.is_valid(), 'Form validated when it shouldn\'t have')
            self.assertEqual(form.errors['blocks'].as_text(), '* B:en:0:html:This field is required')

    def test_create_page_with_nested_blocks(self):
        """ ContainerBlocks are special block types which allow other blocks to be embedded within them """

        block_edit_data = {"es":[],"en":[
            {
                "type": "pageblocks.blocks.ContainerBlock",
                "data": {
                    "class": "col-12",
                    "blocks": [
                        {
                            "type":"pageblocks.blocks.HTMLBlock","data":{"html":"<b>This is a sub block</b>"}
                        }
                    ]
                }
            }
        ]}
        form = PageAdminForm(data={
            'slug': 'basic_page',
            'title': {"es":"prueba", "en":"test"},
            'blocks': block_edit_data
        })
        self.assertTrue(form.is_valid(), form.errors)
        with translation_override('en'):
            page = form.save()

        self.assertEqual(page.slug, 'basic_page')

        self.assertEqual(page.blocks.filter(language='en').count(), 2)
        self.assertEqual(page.blocks.filter(language='en', parent=None).count(), 1)

        block_edit_data['en'][0]['id'] = str(page.blocks.filter(language='en', parent=None).first().id)
        block_edit_data['en'][0]['data']['blocks'][0]['id'] = str(page.blocks.filter(language='en').exclude(parent=None).first().id)

        form = PageAdminForm(instance=page)
        self.assertDictEqual(form.fields['blocks'].initial, block_edit_data)


    # TODO: Test creating a page with a required block field (or type) missing


# Generated by Django 3.2.5 on 2021-10-12 16:50

from django.db import migrations, models
import django.db.models.deletion
import pageblocks.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Page',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('slug', models.SlugField(unique=True)),
                ('title', pageblocks.models.MultiLanguageField(default=dict)),
            ],
        ),
        migrations.CreateModel(
            name='PageBlock',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('language', models.CharField(db_index=True, max_length=2)),
                ('index', models.IntegerField(default=-1)),
                ('type', models.TextField()),
                ('content', models.TextField()),
                ('page', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='blocks', to='pageblocks.page')),
            ],
        ),
    ]



from django.conf import settings
from django.contrib import admin

from .models import Page
from .forms import PageAdminForm


class PageAdmin(admin.ModelAdmin):
    form = PageAdminForm
    change_form_template = 'admin/pageblocks/change_form.html'

    def save_model(self, request, obj, form, change):
        obj.save()
        form.save_blocks(obj)


# admin.site.register(Page, PageAdmin)


import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-page-blocks',
    version='0.3.4',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='A simple, Wagtail CMS inspired content block engine for Django.  Intended to give slightly more control than regular flatpages.',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/bravasoftware/django-page-blocks',
    author='Mark Skelton',
    author_email='studio@bravasoftware.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



from django.http.response import Http404
from django.views.generic.base import TemplateView
from django.utils.translation import gettext


class PageView(TemplateView):
    template_name = None
    queryset = None

    def get_queryset(self):
        if not self.queryset:
            raise Exception(gettext('No queryset provided.  This view must provide either a queryset attribute or get_queryset function'))
        return self.queryset

    def get_object(self, *args, **kwargs):
        slug = self.kwargs.get('slug', None)
        if not slug:
            raise Exception(gettext('Expecting a slug parameter on the url.  You can override this behaviour by overriding the get_object function'))
        
        obj = self.get_queryset().filter(slug=slug).first()
        if not obj:
            raise Http404()

        return obj

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['page'] = self.get_object()
        return ctx